/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.UUID;
import java.util.regex.Pattern;

import com.unitesk.requality.nodetypes.VirtualNode;

public class Attribute implements Comparable<Attribute> {
	private UUID origin;
    private String key;
    private Object value, newvalue;
    private AttributeType type; 
    private AttributeGenerator generator = null;

    public Attribute(UUID node, AttributeType type, String key, Object value, AttributeGenerator gen) {
    	this(node,type,key,null,value,gen);
    }
    public Attribute(UUID node, AttributeType type, String key, Object newvalue, Object value) {
    	this(node,type,key,newvalue,value,null);
    }
    public Attribute(UUID node, AttributeType type, String key, Object newvalue, Object value,AttributeGenerator gen) {
    	this.origin = node;
    	this.type = type;
        this.key = key.replace((char)160, ' ');
        if(type == AttributeType.LIST && !(value instanceof Object[]))
        	value = new Object[]{AttributeType.STRING,value};
        if(type == AttributeType.STRING || type == AttributeType.REFERENCE){
        	if(value!=null)
        		value = value.toString().replace((char)160, ' ');
        	if(newvalue!=null)
        		newvalue = newvalue.toString().replace((char)160, ' ');
        }
        this.value = value;
        this.newvalue = newvalue;
    	this.generator = gen;
    	if(gen!=null){
    		gen.setAttribute(this);
    		this.type = AttributeType.GENERATED;
    		//TODO send type to generator
    	}
    }
    
    public Attribute(TreeNode node, AttributeType type, String key, Object value) {
    	this(node.getUUId(), type, key, value);
    }
    public Attribute(UUID origin, AttributeType type, String key, Object value) {
        this(origin,type,key,null,value,null);
    }

    public String getKey() {
        return key;
    }

    private List toList(Object[] arr){
    	if(arr[0] != AttributeType.LIST)
    		return (List) arr[1];
    	List l = new ArrayList();
    	for(Object o2:(List)(arr[1]))
    		l.add(toList((Object[])o2));
    	return l;
    }
    public Object getValue() {
    	Object o = value;
    	if(newvalue != null)
    		o = newvalue;
    	if(getType().equals(AttributeType.LIST)&&!(o instanceof List))
    		return toList((Object[]) o);
        return o;
    }

    public AttributeType getRawType() {
        return type;
    }

    public AttributeType getType() {
    	if(type==AttributeType.GENERATED && generator!=null)
    		return generator.getAttributeType();
        return type;
    }

    @Override
    public String toString() {
        return "key: "+key+" value: "+(newvalue!=null?newvalue:"null") + " rawValue: "+value + " ("+super.toString()+") gen: "+(getGenerator()==null?"none":getGenerator().toString());
    }

    @SuppressWarnings("unchecked")
    @Override
    public boolean equals(Object obj) {
        if(super.equals(obj)) return true;
        if (!(obj instanceof Attribute)||obj==null) {
            return false;
        }
        Attribute attr = (Attribute) obj;
        if(!key.equals(attr.getKey())) {
            return false;
        }
        if(!type.equals(attr.getRawType())) {
            return false;
        }
        if (type == AttributeType.LIST && value instanceof List && attr.getRawValue() instanceof List) {
            List<Object> list1 = (List<Object>)value;
            List<Object> list2 = new ArrayList<Object>((List<Object>)attr.getValue());
            for (Object str : list1) {
                if (list2.contains(str)) {
                    list2.remove(str);
                } else {
                    return false;
                }
            }
            if (!list2.isEmpty()) {
                return false;
            }
        }
        if(generator==null && attr.generator!=null || generator!=null && attr.generator==null || attr.generator!=null && generator != null 
        		&& !attr.generator.getUUID().equals(generator.getUUID()))
        	return false;
        if(value==null&&attr.getValue()!=null || value!=null && !valEquals(getValue(),attr.getValue())) {
            return false;
        }
        return true;
    }
    public boolean weakEquals(Object obj) {
        if(super.equals(obj)) return true;
        if (!(obj instanceof Attribute)||obj==null) {
            return false;
        }
        Attribute attr = (Attribute) obj;
        if(!key.equals(attr.getKey())) {
            return false;
        }
        if(!this.getType().equals(attr.getType())) {
            return false;
        }
        if (getType() == AttributeType.LIST && getValue() instanceof List && attr.getValue() instanceof List) {
            List<Object> list1 = (List<Object>)getValue();
            List<Object> list2 = new ArrayList<Object>((List<Object>)attr.getValue());
            for (Object str : list1) {
                if (list2.contains(str)) {
                    list2.remove(str);
                } else {
                    return false;
                }
            }
            if (!list2.isEmpty()) {
                return false;
            }
        }
        if(getValue()==null&&attr.getValue()!=null || getValue()!=null && !valEquals(getValue(),attr.getValue())) {
            return false;
        }
        return true;
    }
    private static boolean valEquals(Object value2, Object object) {
    	if (value2 instanceof Object[] && object instanceof Object[]){
    		Object[]o1=(Object[])value2,o2=(Object[])object;
    		if(o1.length == o2.length && o1.length == 2){
    			if(!o1[0].equals(o2[0]))
    				return false;
    			o1 = (Object[]) o1[1];
    			o2 = (Object[]) o2[1];
    			return Arrays.equals(o1, o2);
    		}else
    			return false;
    	}
		return value2.equals(object);
	}
	static final Pattern valid_inner_name_pattern = Pattern.compile("[A-Za-z0-9а-яА-Я_[$].-]+[A-Za-z0-9а-яА-Я._[$] -]*");
    static final Pattern valid_name_pattern = Pattern.compile("[A-Za-z0-9а-яА-Я.-]+[A-Za-z0-9а-яА-Я._[$] -]*");
    public static boolean checkNameIsValid(String name, boolean inner){
        return inner ? valid_inner_name_pattern.matcher(name).matches() : valid_name_pattern.matcher(name).matches();
    }

    @Override
    public int compareTo(Attribute o) {
        return this.getKey().compareTo(o.getKey());
    }

    @SuppressWarnings("unchecked")

    public static Attribute deepCopy(Attribute attr) {
    	return deepCopy(attr,null);
    }
    public static Attribute deepCopy(Attribute attr, UUID newuid) {
    	if(attr == null)
    		return null;
        Object value = attr.getValue(), rawval = attr.getRawValue();
        if (attr.getType() == AttributeType.LIST && value instanceof List) {
        	Object[]o = (Object[])rawval;
            List val = new ArrayList((List)o[1]);
            //TODO proper generator handling
        	return new Attribute(newuid!=null?newuid:attr.origin, AttributeType.LIST, attr.getKey(),null, new Object[]{o[0],val}, attr.generator!=null?attr.generator.copy():null);
        }
        return new Attribute(newuid!=null?newuid:attr.origin, attr.type, attr.key, null,attr.getRawValue(), attr.generator!=null?attr.generator.copy():null);
    }

    public AttributeType getListType(){
    	if(getRawType().equals(AttributeType.GENERATED)&&getGenerator()!=null)
    		return getGenerator().getAttributeListType();
    	if(getRawType().equals(AttributeType.LIST))
    		return (AttributeType) ((Object[])value)[0];
    	return null;
    }
	public Object getRawValue() {
		return value;
	}

	public UUID getOrigin() {
		return origin;
	}
	
	public boolean isGenerated(){
		return getRawType().equals(AttributeType.GENERATED)&&generator!=null;
	}

	public boolean isMoreVisible(TreeDB db, Attribute a) {
		if(isGenerated()){
			if(a.isGenerated()){
				AttributeGenerator ag1 = generator, ag2 = a.generator;
				if(!db.isVirtual()){
					if(ag1==null)
						ag1=VirtualNode.getVirtualTreeDB(db).getNode(getOrigin()).getAttribute(getKey()).generator;
					if(ag2==null)
						ag2=VirtualNode.getVirtualTreeDB(db).getNode(a.getOrigin()).getAttribute(a.getKey()).generator;
				}
				if(ag1==null || ag2==null)
					return false;
				return ag1.moreVisibleThan(ag2);
			}
			else 
				return false;
		}else
			return true;
	}
	public static String valueToString(AttributeType at, Object value){
		if(at == AttributeType.LIST){
			if(!(value instanceof Object[]))
				value = new Object[]{AttributeType.STRING,value};
			Object[]o = (Object[])value;
			StringBuilder sb = new StringBuilder();
			AttributeType at2 = (AttributeType) o[0];
			sb.append(at2);
			sb.append("=[");
			if(at2==AttributeType.LIST){
				java.util.Iterator it = ((List)o[1]).iterator();
				while(it.hasNext()){
					Object oi=it.next();
					sb.append(valueToString(at2,oi));
					if(it.hasNext())
						sb.append(",");
				}
			}else{
				java.util.Iterator it = ((List)o[1]).iterator();
				while(it.hasNext()){
					Object oi=it.next();
					sb.append(oi.toString());
					if(it.hasNext())
						sb.append(",");
				}
			}
			sb.append("]");
    		return sb.toString();
		}
		return value!=null?value.toString():"";
	}
	public void setValue(Object object) {
		value = object;
		newvalue = null;
	}
	public void setNewValue(Object object) {
		newvalue = object;
	}
	public AttributeGenerator getGenerator() {
		if(getRawType().equals(AttributeType.GENERATED))
			return generator;
		return null;
	}
	public void setGenerator(AttributeGenerator ag) {
		generator = ag;
	}

}
