/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import com.unitesk.requality.nodetypes.VirtualNode;

public class NodeDesc {
	//taken from transaction storage
    private static final String TMP_RESOURCE_PREFIX = "requality";
	private UUID uuid;
    private String id;
    private Map<String, Attribute> attributes;
    private Map<String, File> resources;
    private List<UUID> childrenUUId;
    private Map<UUID, NodeDesc> children;
    private UUID parentUUId;

    public NodeDesc(UUID uuid, UUID parentUUId, String id,
            Map<String, Attribute> attributes, List<UUID> childrenUUId) {
        this.uuid = uuid;
        this.id = id;
        this.attributes = new HashMap<String, Attribute>();
        for (String key : attributes.keySet()) {
            Attribute attr = attributes.get(key);
            if(!attr.isGenerated()){
            	Object value = attr.getRawValue();
            	if (value instanceof Object[]) {
                	value = renewList((Object[])value,(List<?>)attr.getValue());
            	}
            	this.attributes.put(key, new Attribute(uuid, attr.getType(), attr.getKey(), value, attr.getGenerator()));
            }else
            	this.attributes.put(key, new Attribute(uuid, attr.getRawType(), attr.getKey(), attr.getRawValue(), attr.getGenerator()));
        }
        this.childrenUUId = new ArrayList<UUID>(childrenUUId);
        this.children = new HashMap<UUID, NodeDesc>();
        this.parentUUId = parentUUId;
    }

    private Object renewList(Object[] value, List<?> newValues) {
    	if(value[0] == AttributeType.LIST){
    		List<Object> k = new ArrayList<Object>();
    		Iterator<?> i = newValues.iterator();
    		for(Object o: (List<?>)value[1]){
    			Object o2=i.next();
    			k.add(renewList((Object[])o, (List)o2));
    		}
    		return new Object[]{value[0],k};
    	}else{
    		return new Object[]{value[0],new ArrayList(newValues)};
    	}
	}

	public void setId(String id) {
        this.id = id;
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof NodeDesc)) {
            return false;
        }
        NodeDesc nodeDesc = (NodeDesc) obj;
        if (!nodeDesc.getId().equals(getId())) {
            return false;
        }
        if (!nodeDesc.getUUId().equals(getUUId())) {
            return false;
        }
        if (!attrubuteEquals(nodeDesc)) {
            return false;
        }
        if (!childrenEquals(nodeDesc)) {
            return false;
        }
        return true;
    }

    public boolean equalsButUUID(Object obj) {
        if (!(obj instanceof NodeDesc)) {
            return false;
        }
        NodeDesc nodeDesc = (NodeDesc) obj;
        if (!nodeDesc.getId().equals(getId())) {
            return false;
        }
        if (!attrubuteEquals(nodeDesc)) {
            return false;
        }
        if (!childrenEquals(nodeDesc)) {
            return false;
        }
        return true;
    }

    private boolean childrenEquals(NodeDesc nodeDesc) {
        return getChildrenUUId().equals(nodeDesc.getChildrenUUId());
    }

    private boolean attrubuteEquals(NodeDesc nodeDesc) {
        return nodeDesc.getAttributes().equals(getAttributes());
    }

    public String getId() {
        return id;
    }

    public UUID getUUId() {
        return uuid;
    }

    public List<UUID> getChildrenUUId() {
        return Collections.unmodifiableList(childrenUUId);
    }

    public Map<String, Attribute> getAttributes() {
        return attributes;
    }

    public void setAttribute(Attribute attr) {
        attributes.put(attr.getKey(), attr);
    }

    public Attribute removeAttribute(String key) {
        return attributes.remove(key);
    }

    public UUID getParentUUId() {
        return parentUUId;
    }

    public String getType() {
        return getAttributes().get(TreeNode.ATTR_TYPE).getValue().toString();
    }

    public void setChildrenDesc(UUID chUUId, NodeDesc chDesc) {
        assert childrenUUId.contains(chUUId);
        children.put(chUUId, chDesc);
    }

    public NodeDesc getChildDesc(UUID chUUId) {
        return children.get(chUUId);
    }
    public TreeNode createNode(TreeDB db, UUID pUUId, Boolean deep) {
    	return createNode(db, pUUId, deep, false);
    }
    	   
    public TreeNode createNode(TreeDB db, UUID pUUId, Boolean deep, Boolean saveUUID) {
        // new node - new uuid
    	UUID uid = uuid;
    	if(!saveUUID)
    		uid = db.getUuidGenerator().nextUUId();
        TreeNode result = db.createChildNode(uid,
                pUUId, id, getType());
        for (Attribute a : getAttributes().values()) {
            result.putAttribute(a);
        }
        result.saveAttributes();
        if (deep) {
            for (UUID chUUId : getChildrenUUId()) {
                children.get(chUUId).createNode(db, result.getUUId(), deep,saveUUID);
            }
        }
        if(resources!=null)
        	for (String s:resources.keySet())
        		try {
        			File f = resources.get(s);
        			if(f.exists())
        				result.setResourceContent(s, new FileInputStream(f));
        		} catch (FileNotFoundException e) {
        			e.printStackTrace();
        		}
        return result;
    }
    public TreeNode createNode(TreeDB db) {
    	return createNode(db,false);
    }
    public TreeNode createNode(TreeDB db, Boolean saveLink) {
        return createNode(db, parentUUId, true,saveLink);
    }

    public void setParentUUId(UUID pUUId) {
        this.parentUUId = pUUId;
    }

    static public NodeDesc fromNode(TreeNode node) {
        return fromNode(node, true);
    }

    static public NodeDesc fromNode(TreeNode node, Boolean deep) {
    	List<UUID> uids = node.getRawChildrenUUIDs();
    	if(node.getType().equals(VirtualNode.TYPE_NAME))
    		uids = Collections.EMPTY_LIST;
        NodeDesc result = new NodeDesc(node.getUUId(), node.getParentUUId(true),
                node.getId(), node.getAttributes(), uids);
        String[]ress = node.listResources();
        result.resources = new HashMap<String, File>();
        for(String s : ress){
        	InputStream is = node.getResourceContent(s);
        	if(is == null)
        		continue;
        	byte[] bt;
			try {
				File tmpNewFile = File.createTempFile(TMP_RESOURCE_PREFIX, node.getUUId().toString() + result.resources.size());
				bt = new byte[is.available()];
	        	is.read(bt);
				FileOutputStream fos = new FileOutputStream(tmpNewFile);
				fos.write(bt);
				fos.close();
				result.resources.put(s, tmpNewFile);
				tmpNewFile.deleteOnExit();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
        }
 //       node.getResourceContent(name)
        if (deep) {
            int num = node.sizeChildren();
            for (int i = 0; i < num; i++) {
                TreeNode child = node.getChild(i);
                result.setChildrenDesc(child.getUUId(), child.toNodeDesc());
            }
        }
        return result;
    }

	public void setChildrenUUIDs(List<UUID> childs) {
		childrenUUId = childs;
	}
}