/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.documents;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.xml.sax.SAXException;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.views.documents.DiffProcessor.Diff;
import com.unitesk.requality.eclipse.views.documents.DiffProcessor.Operation;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.NodeType;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.ToDoItem;
import com.unitesk.requality.tools.RequalityCLI;

/**
 * Handles the action in the context menu.  
 * Transfer requirements from old document to new document 
 * 
 */
public class ProcessNewSpecificationVersion extends DocumentProcessor {
    private Document oldDocument = null;
    private Document newDocument = null;
    private TreeDB db;
    
	private int diffFoundLocCounter = 0;
	private int transferedLocCounter = 0;
	
	private DocStringRepresentation oldDocRep = null;
	private DocStringRepresentation newDocRep = null;
	
	private int outsideEqualCount = 0;
	private int partiallyOutsideEqual = 0;
	
	private int completeReqsCount = 0;

	private int incompleteReqsCount = 0;
	private int failedReqsCount = 0;
	private boolean showMessageBoxWithStatistics = true;
	
	public ToDoList toDoList;
	
	
	private Map<String, Integer> reqIdsToLocCounter = new HashMap<String, Integer>();
	private Map<String, Integer> reqIdsToTransferredLocCounter = new HashMap<String, Integer>();
	
	
	public void unsetShowMessageBoxWithStats() {
		showMessageBoxWithStatistics = false;
	}
	
	public void setShowMessageBoxWithStats() {
		showMessageBoxWithStatistics = true;
	}
	
	public boolean verbose = false;
	public boolean writeDebugFiles = false;
	
    @Override
    public boolean configure(final Document doc) {
    	return configure(doc, null);
    }

    public boolean configure(Document doc, Document prefferedDoc, boolean force) {
        if (!super.configure(doc)) {
            return false;
        }
        this.db = doc.getTreeDB();
    	
    	if (doc.getAttribute(Document.ATTR_ALREADY_UPDATED) != null ) {
    		super.showMessageBox("Document Processor","This document was already updated. "
    				+ "Documents cannot be updated more than once.");
			return false;
		}
    	
        oldDocument = showSelectAnotherDocumentDialog(doc, prefferedDoc, force);
        return oldDocument!=null;
    }
    
    /**
     * This method preferred to default. Can be used in console mode.
     */
    public boolean configure(Document doc, Document prefferedDoc) {
    	
    	return configure(doc,prefferedDoc,false);
    }
    
    
    @Override
    protected boolean process(String docId, final Reader in, 
                              final Writer out, TreeNode rootNode) {
 
    	newDocument = (Document) db.getNode(docId);
    	if(isCanceled())
			return false;
    	return processReqsTransfer(out);
    	
    }
    
    
    /** 
     * Methods transfers locations from old document to new document. Can be used in console mode
     */
    public void transferRequirements(Document oldDoc, Document newDoc) {
    	
    	newDocument = newDoc;
    	oldDocument = oldDoc;
    	
    	this.db = oldDocument.getTreeDB();
    	
		ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
		Writer out = null;
        try {
			out = new OutputStreamWriter(byteOut, encoding);
			processReqsTransfer(out);
	        out.close();
	        String htmlName = newDocument.getResourceName();
	        
	        if (verbose) {
	        System.out.println("htmlName = " + htmlName);
	        }
	        
	        ByteArrayInputStream outIn = new ByteArrayInputStream(byteOut.toByteArray());
	        newDocument.setResourceContent(htmlName, outIn, new ProcessorData());
            outIn.close();
	        
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
        
    }
    
    private boolean processReqsTransfer(final Writer out) throws TransformerFactoryConfigurationError {
    	transferedLocCounter = 0;
		diffFoundLocCounter = 0;
		
		completeReqsCount = 0;
		incompleteReqsCount = 0;
		failedReqsCount = 0;
		
    	try {
    		if(isCanceled())
				return false;
    		
	    	newDocRep = new DocStringRepresentation(newDocument.getDOM());
	    	
	    	if(isCanceled())
				return false;
	    	
	    	oldDocRep = new DocStringRepresentation(oldDocument.getDOM());
	    	
	    	if (writeDebugFiles) {
    			printCleanStringsToFiles();
	    	}
	    	
	    	if(isCanceled())
				return false;
	    	
	    	Map<SpanLocation, List<LocationToTransfer>> locationsInNewFile = processDiff();
	    	
	    	if(isCanceled())
				return false;
			
	    	boolean status = updateReqAttributes(locationsInNewFile);
	    	
	    	if (!status) {
	    		return false;
	    	}
	    	if(isCanceled())
				return false;
	    	
	    	status = setStatuses();
	    	
	    	if (!status) {
	    		return false;
	    	}
	    	
	    	
	    	
	    	if(isCanceled())
				return false;
	    	
	    	
	    	
			newDocRep.deleteRemovedNodes();
			
			status = createToDoLinks();
			
			if (verbose) {
				System.out.println("ToDo list items created");
			}
			
			if (!status) {
	    		return false;
	    	}
	    	
	    	
	    	if(isCanceled())
				return false;
	    	
			
			newDocument.setXmlDoc(newDocRep.doc);
			newDocument.saveDOMFromXMLCache();
			
			if(isCanceled())
				return false;
			
			transform(out);
			
			if(isCanceled())
				return false;
			
			oldDocument.addUpdatedProblem();
			newDocument.addUpdatedProblem();
			
			newDocument.setXmlDoc(newDocRep.doc);
			newDocument.saveDOMFromXMLCache();
			
			newDocument.markDocumentAsUpdated();
			
			if (verbose) {
				System.out.println("newDocument.saveDOMFromXMLCache() is done");
			}
			
			
			PlatformUI.getWorkbench().getDisplay().syncExec(new Runnable() {
	    	    public void run() {
	    	    			if (verbose) {
								System.out.println("In run() for ToDoList view");
							}
//	    	    				 
							
							try {
								ToDoList viewPart;
								viewPart = (ToDoList) PlatformUI.getWorkbench()
										.getActiveWorkbenchWindow().getActivePage().showView("com.unitesk.requality.views.ToDoList");
								viewPart.refresh(db);
							} catch (PartInitException e) {
								e.printStackTrace();
							}
	    	    }
			});
			
			generateTextForMessageBox();
			
			return true;
	       
		} catch (TransformerConfigurationException e) {
			e.printStackTrace();
		} catch (TransformerException e) {
			e.printStackTrace();
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		} 
    	
    	return false;
    }

	private void printCleanStringsToFiles() throws FileNotFoundException {
		PrintWriter oldCleanString = new PrintWriter(new File("oldCleanStr.txt"));
//    		oldCleanString.print(oldDocRep.getCleanString().replaceAll("\n", "@"));
		oldCleanString.print(oldDocRep.getCleanString());
		oldCleanString.close();
		
		PrintWriter oldTagsFile = new PrintWriter(new File("oldTags.txt"));
		for (NodeLocation nodeLoc : oldDocRep.getTags())
			oldTagsFile.println(nodeLoc);
		oldTagsFile.close();
		
		PrintWriter newCleanString = new PrintWriter(new File("newCleanStr.txt"));
//	    	newCleanString.print(newDocRep.getCleanString().replaceAll("\n", "@"));
		newCleanString.print(newDocRep.getCleanString());
		newCleanString.close();
		PrintWriter newTagsFile = new PrintWriter(new File("newTags.txt"));
		for (NodeLocation nodeLoc : newDocRep.getTags())
			newTagsFile.println(nodeLoc);
		newTagsFile.close();
	}
    
    private boolean updateReqAttributes(Map<SpanLocation, List<LocationToTransfer>>  locationsInNewFile) {
		
		for (SpanLocation spanLoc : oldDocRep.getSpans()) {
			String id = spanLoc.getSpanId().substring(18);
			if (reqIdsToLocCounter.keySet().contains(id)) {
				int oldCount = reqIdsToLocCounter.get(id);
				reqIdsToLocCounter.put(id, oldCount + 1);
			} else {
				reqIdsToLocCounter.put(id, 1);
			}
		}
		
		if(isCanceled())
			return false;
		
		Map<String,UUID> idToUUID = new HashMap<String, UUID>();
		for (SpanLocation spanLoc : oldDocRep.getSpans()) {
			
			if (verbose) {
				System.out.println("spanLoc = " + spanLoc);
			}
			
			if(isCanceled())
				return false;
			
			List<LocationToTransfer> locsToTransfer = locationsInNewFile.get(spanLoc);
			if (locsToTransfer != null) {
				// currently there is max one element in locsToTransfer, this cycle is for 
				// preposition of different variants to user
				for (LocationToTransfer locToTransfer : locsToTransfer) {
					++diffFoundLocCounter;
					boolean forUpdate = false;
					if(!idToUUID.containsKey(spanLoc.getSpanId())) {
						idToUUID.put(spanLoc.getSpanId(), UUID.randomUUID());
					}
					UUID spanID = idToUUID.get(spanLoc.getSpanId());
					if (locToTransfer.getSpanText().equals(spanLoc.getText())) {
						forUpdate = newDocRep.DocModify(locToTransfer, "requality_text id_" + spanID.toString());
					} else {
						if (verbose ) {
							System.out.println("Different span texts to modify. Continue.");
						}
						
					}
					
					if (forUpdate) {
						// adding locations in new doc for current requirement
						++transferedLocCounter;
						String sspanID = spanID.toString();
						String oldspanID = spanLoc.getSpanId().substring(18);
						Location oldLoc = oldDocument.getLocation(oldspanID);
						
						if (reqIdsToTransferredLocCounter.keySet().contains(oldspanID.toString())) {
							int count = reqIdsToTransferredLocCounter.get(oldspanID.toString());
							reqIdsToTransferredLocCounter.put(oldspanID.toString(), count + 1);
						} else {
							reqIdsToTransferredLocCounter.put(oldspanID.toString(), 1);
						}

						Location loc;
						
						if (newDocument.getLocation(sspanID) == null ) {
							loc = newDocument.createLocation(spanID);
							loc.saveAttributes();
							if (oldLoc != null) {
						       	Requirement[] reqs = oldLoc.getRequirements();
					            for (Requirement req : reqs) {
					               req.addLocation(loc.getQualifiedId());
					               req.saveAttributes();
					            }
					        }
						} 
				        
				       
					}
				}
			}
		}

		
		 return true;
	}
    
    
    private boolean setStatuses() {
    	
    	
    	if (verbose) {
    		System.out.println("reqIdsToTransferredLocCounter:");
    		for (String str : reqIdsToTransferredLocCounter.keySet()) {
    			System.out.println(str + " " + reqIdsToTransferredLocCounter.get(str));
    		}
    	}
    	

		 
    	Set<Requirement> requirementsFromOldDoc = new HashSet<Requirement>();
    	
    	if (verbose) {
			System.out.println("in setStatuses " );
		}
    	
    	
    	for (UUID uuid : oldDocument.getLocations()) {
    		
    		if (verbose) {
    			System.out.println("uuid =  " + uuid);
    		}
			 
    		Location location =  (Location) oldDocument.getLocation(uuid.toString());
    		
    		
    		
    		if (location == null) {
    			// TODO add checkers before update processor
    			continue;
    		}
    		
    		if (verbose) {
    			System.out.println("location = " + location);
    		}
    		
    		for (Requirement req : location.getRequirements() ) {
    			requirementsFromOldDoc.add(req);
    		}
    	}
    		
		 if(isCanceled()) {
				return false;
		 }
			 
		 for (Requirement req : requirementsFromOldDoc) {
			 // uncomment if we don't want to mark structure requirements without locations with statuses overlays   
//    			if (req.getLocations().isEmpty()) {
//    			  continue;
//    			}
			 
			 if (verbose) {
				 System.out.println("req = " + req);
			 }
				 
			 List<Location> reqLocations = req.getLocations();
			 int numberOfTransferredSpans = 0;
			 int numberOfInitialSpans = 0;
			 
			 for (Location loc : reqLocations) {
				 
				 
				 if (verbose) {
					 System.out.println("loc = " + loc + " loc.getId() = " + loc.getId());
				 }
				 
				 if (reqIdsToTransferredLocCounter.keySet().contains(loc.getId())) {
					 numberOfTransferredSpans += reqIdsToTransferredLocCounter.get(loc.getId());
				 }
				 if (reqIdsToLocCounter.keySet().contains(loc.getId())) {
					 numberOfInitialSpans += reqIdsToLocCounter.get(loc.getId());
				 }
			 }
			 
			 Requirement.TransferStatus statusOfCurrentReq = null;
			 
			 if (numberOfInitialSpans > 0) {
				 if (numberOfTransferredSpans == numberOfInitialSpans) {
					 statusOfCurrentReq = Requirement.TransferStatus.COMPLETE;
					 completeReqsCount++;
				 } 
				 if ( (numberOfTransferredSpans > 0) && (numberOfTransferredSpans < numberOfInitialSpans) ) {
					 statusOfCurrentReq = Requirement.TransferStatus.INCOMPLETE;
					 incompleteReqsCount++;
				 }
				 if (numberOfTransferredSpans == 0) {
					 statusOfCurrentReq = Requirement.TransferStatus.FAILED;
					 failedReqsCount++;
				 }
			 }
			 if (statusOfCurrentReq != null) {
				 req.setTransferProblem(statusOfCurrentReq);
			 }
			 
			 if (verbose) {
				 System.out.println("numberOfTransferredSpans = " + numberOfTransferredSpans);
				 System.out.println("numberOfInitialSpans = " + numberOfInitialSpans);
				 System.out.println("status = " + statusOfCurrentReq);
			 }
		 }
	
		 
		 for (TreeNode node : getOrderFromLeaves(db.getRootNode())) {
			 
			 if (verbose) {
				 System.out.println("node = " + node);
			 }
			 
			 if(isCanceled()) {
					return false;
			 }
			 
			 
			 if (node instanceof Requirement) {
				 Requirement req = (Requirement) node;
				 Requirement.TransferStatus worstStatus = null;
				 
				 Attribute reqAttr = req.getAttribute(Requirement.ATTR_TRANSF_STATUS);
				 if (reqAttr != null) {
					 worstStatus = Requirement.TransferStatus.valueOf(reqAttr.getRawValue().toString());
				 }
				 
				 
				 if (node.sizeChildren() > 0) {
					 TreeNode[] children = node.getSortedChildren();
					 for (TreeNode child : children) {
						 Attribute attr = child.getAttribute(Requirement.ATTR_TRANSF_STATUS);
						 if (attr != null) {
							 Requirement.TransferStatus childStatus = Requirement.TransferStatus.valueOf(attr.getRawValue().toString());
							 if (childStatus != null) {
								 if (worstStatus == null ) {
									 worstStatus = childStatus; 
								 } else if (!childStatus.isBetterThan(worstStatus)) {
									 worstStatus = childStatus;
								 }
							 }
						 }
					 }
				 }
				 
				 if (worstStatus != null) {
					 req.setTransferProblem(worstStatus);
				 }
			 }
		 }
		 return true;
    }
    
    
    
    private boolean createToDoLinks() {
    	TreeNode toDoFolder = db.getNode("/ToDoLinks",true);
    	if (toDoFolder == null) {
    		toDoFolder = db.getRootNode().createChild("ToDoLinks", NodeType.TYPE_NAME);
    	}
    	
    	UUID root = db.getNode(Requirement.getTypeRootQId()).getUUId();
    	for (TreeNode node : getOrderFromLeaves(db.getNode(Requirement.getTypeRootQId()))) {
			 
    		if (node.getUUId().equals(root)) {
    			continue;
    		}
    		
			 if (verbose) {
				 System.out.println("node = " + node);
			 }
			 
			 if(isCanceled()) {
					return false;
			 }
			 
			 
			 if (node instanceof Requirement) {
				 Requirement req = (Requirement) node;
				 
				 Requirement.TransferStatus status = null;
				 
				 Attribute reqAttr = req.getAttribute(Requirement.ATTR_TRANSF_STATUS);
				 if (reqAttr != null) {
					 status = Requirement.TransferStatus.valueOf(reqAttr.getRawValue().toString());
				 }
				 
				 if (status != null) {
					 TreeNode toDoItem = toDoFolder.createChild(RequalityCLI.getNextId(toDoFolder, ToDoItem.TYPE_NAME, "link_"), ToDoItem.TYPE_NAME);
					 Task.Action action; 
					 switch (status) {
					 	case INCOMPLETE:
					 		action = Task.Action.ADD_LOCATIONS;
					 		break;
					 	case COMPLETE:
					 		action = Task.Action.VERIFY;
					 		break;
					 	case FAILED:
					 		action = Task.Action.FIND_LOCATIONS;
					 		break;
				 		default: 
				 			action = Task.Action.FIND_LOCATIONS;
				 			break;
		 			}
					 
			    	 toDoItem.putAttribute(new Attribute(toDoItem, AttributeType.STRING, ToDoItem.ATTR_REQ_LINK, req.getUUId()));
			    	 toDoItem.putAttribute(new Attribute(toDoItem, AttributeType.STRING, ToDoItem.ATTR_TRANSF_ACTION, action));
			    	 toDoItem.putAttribute(new Attribute(toDoItem, AttributeType.STRING, ToDoItem.OLD_DOC, oldDocument.getQualifiedId()));
			    	 toDoItem.putAttribute(new Attribute(toDoItem, AttributeType.STRING, ToDoItem.NEW_DOC, newDocument.getQualifiedId()));
			    	 toDoItem.saveAttributes();
				 }
			 }
		 }
    	if (verbose) {
    		System.out.println("Before exiting ToDoLinks creation");
    	}
    	return true;
    }
    

	private List<TreeNode> getOrderFromLeaves(TreeNode rootNode) {
		List<TreeNode> order = new ArrayList<TreeNode>();
		bfs(rootNode, order);
		Collections.reverse(order);
		return order;
	}
	
	private void bfs(TreeNode node, List<TreeNode> queue) {
		queue.add(node);
		for (TreeNode child : node) {
			bfs(child, queue);
		}
	}

	private Map<SpanLocation, List<LocationToTransfer>> processDiff() {

		Map<SpanLocation, List<LocationToTransfer>> locationsInNewFile = new HashMap<SpanLocation, List<LocationToTransfer>>();
		
		DiffProcessor differ = new DiffProcessor();
		LinkedList<Diff> diffResult;
		
		diffResult = differ.diff_main(oldDocRep.getCleanString(), newDocRep.getCleanString(), true);
		
		int indexInDiffList = 0;
	
		for (SpanLocation spanLoc : oldDocRep.getSpans()) {
			String spanText = spanLoc.getText();
			int spanBeginIndex = spanLoc.getLocation().getStart();
			Diff df = diffResult.get(indexInDiffList);
			boolean isSpanStartInsideDiffBlock = 
					(spanBeginIndex >= ( df.getOriginalOffset() - (df.getText().length()))) &&
					(spanBeginIndex <= df.getOriginalOffset() );
			while (!isSpanStartInsideDiffBlock) {
				++indexInDiffList;
				df = diffResult.get(indexInDiffList);
				isSpanStartInsideDiffBlock = 
						(spanBeginIndex >= ( df.getOriginalOffset() - (df.getText().length()))) &&
						(spanBeginIndex <= df.getOriginalOffset() );
			}
			List<LocationToTransfer> locationsToTransfer = new ArrayList<LocationToTransfer>();
			Diff currentDiff = df;
			if (currentDiff.getOperation().equals(Operation.EQUAL)) {
				int spanTextLen = spanText.length();
				if ( (spanBeginIndex + spanTextLen) <= 
						(currentDiff.getOriginalOffset() ) ) {
					// the span is all inside EQUAL diff
					int startIndexInNewDoc = currentDiff.getReplacementOffset() + spanBeginIndex - currentDiff.getOriginalOffset();
					LocationToTransfer locToTransfer = new LocationToTransfer(startIndexInNewDoc, 
							startIndexInNewDoc + spanTextLen, spanText);
					locationsToTransfer.add(locToTransfer);
					locationsInNewFile.put(spanLoc, locationsToTransfer);
					
				} else {
					// TODO upgrade for case of partially in equal block
					++partiallyOutsideEqual;
				}
			} else {
				++outsideEqualCount;
			}
		}
		
		return locationsInNewFile;
	}

	public int getOutsideEqualCount() {
		return outsideEqualCount;
	}

	public int getPartiallyOutsideEqual() {
		return partiallyOutsideEqual;
	}

	private void generateTextForMessageBox() {
		
		long totalReqCounter = completeReqsCount + incompleteReqsCount + failedReqsCount;
		
		long transferredCounter = completeReqsCount + incompleteReqsCount;

		String message = "Document processing finished.\n";
		
		if (totalReqCounter == 0) {
			message += "There is no extracted requirements in marked document";
		} else {
		
		
			if (transferredCounter == 0) {
				message += "None of the requirements have been transferred from " + totalReqCounter + " requirements.";
			} else {
				message += "Transferred " + transferredCounter + " requirement";
				if (transferredCounter > 1) {
					message += "s";
				}
				message += " from " + totalReqCounter + ".";
				
				if (incompleteReqsCount > 0) {
					message += "\nWhere for " + incompleteReqsCount + " requirement";
					if (incompleteReqsCount > 1) {
						message += "s";
					}
					message += " not all text fragments have been transferred.";
				}
			}
		}
			
		
		super.messageBoxMessage = message;
		super.messageBoxTitle = "Document Processor";
		
		if (verbose) {
			System.out.println("From MessageBox: " + message);
			
			System.out.println("Stats:");
			System.out.println("number of spans: " + oldDocRep.getSpans().size());
			System.out.println("diff found: " + diffFoundLocCounter + " spans to transfer");
			System.out.println("number of transferred spans: " + transferedLocCounter);
			System.out.println("outsideEqualCount: " + outsideEqualCount);
			System.out.println("partiallyOutsideEqual: " + partiallyOutsideEqual);
			
			System.out.println("Partially transferred number of reqs : " + incompleteReqsCount);
			System.out.println("Totally transferred number of reqs: " + completeReqsCount);
			System.out.println("Not transferred number of reqs: " + failedReqsCount);
			System.out.println("Total number of reqs in initial doc: " + totalReqCounter);
		}
		
		
		
	}

	private void transform(final Writer out) throws Exception,
			TransformerFactoryConfigurationError,
			TransformerConfigurationException, TransformerException {
		DOMSource domSource = new DOMSource(newDocument.getDOM());

		StreamResult result = new StreamResult(out);
		TransformerFactory tf = TransformerFactory.newInstance();
		Transformer transformer;

		transformer = tf.newTransformer();
		transformer.setOutputProperty(OutputKeys.INDENT, "no");
		transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
		transformer.transform(domSource, result);
		
	}
    
    
	public int getTransferedLocCounter() {
		return transferedLocCounter;
	}
	
	public int getDiffFoundLocCounter() {
		return diffFoundLocCounter;
	}
	
	public int getTotalSpansCounter() {
		return oldDocRep.getSpans().size();
	}
	
	public int getCompleteReqsCount() {
		return completeReqsCount;
	}

	public int getIncompleteReqsCount() {
		return incompleteReqsCount;
	}

	public int getFailedReqsCount() {
		return failedReqsCount;
	}
	
	public int getTotalReqsCount() {
		return (completeReqsCount + incompleteReqsCount + failedReqsCount);
	}
    
}
