/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeComparator;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.RequalityCLI;
public class Location extends TreeNode {
    public static final String TYPE_NAME = "Location";
    private String label;
//    private boolean isTransferred = false;

	public Location() {
        super(null, null);
    }

    public Location(TreeDB db, UUID uid) {
        super(db, uid);
        setType(TYPE_NAME);
    }
    private String getHTMLFolder(){
    	TreeNode parent = getParent();
    	//in a case of tests
    	if(parent == null)
    		return "";
    	return new File(
                ((com.unitesk.requality.nodetypes.Document) parent )
                        .getHTMLPath()).getParent();
    }
    
    public Object getDocumentNodeModel(){
    	Document d = getDocument();
    	if(d == null)
    		return null;
    	return d.getNodeModel();
    }
    
    @Override
    public String getDescription(String prefix, String postfix, String basePath) {
    	String desc = getDescription();
    	if(desc == null)
    		desc = "";
        return CoreUtils.htmlEncapsulate(RequalityCLI.getReportDescription(this, getHTMLFolder(),
                desc, prefix, postfix, basePath));
    }

    public Requirement[] getRequirements() {
        String locQId = getQualifiedId();
        return getRequirementsFor(getTreeDB(), locQId);
    }

    private static Map<TreeDB,Map<String,Set<Requirement>>> reqCahceMap = new HashMap<TreeDB,Map<String,Set<Requirement>>>();
    public static Requirement[] getRequirementsFor(TreeDB db, String locQId) {
    	if(!reqCahceMap.containsKey(db))
    		reqCahceMap.put(db, new HashMap<String, Set<Requirement>>());
    	if(reqCahceMap.get(db).containsKey(locQId)){
    		Requirement[]reqs = reqCahceMap.get(db).get(locQId).toArray(new Requirement[0]);
    		Arrays.sort(reqs,reqcomparator);
    		return reqs;
    	}
        Set<Requirement> result = new HashSet<Requirement>();
        for (TreeNode node : new DeepFirstTreeWalker(db.getNode(Requirement.getTypeRootQId()))) {
            if (!node.getType().equals(Requirement.TYPE_NAME)) {
                continue;
            }
            Requirement req = (Requirement) node;
            for(String lid:req.getLocationQIds()){
            	if(!reqCahceMap.get(db).containsKey(lid))
            		reqCahceMap.get(db).put(lid, new HashSet<Requirement>());
            	reqCahceMap.get(db).get(lid).add(req);
            	if(lid.equals(locQId))
            		result.add(req);
            }
        }
        Requirement[] res = result.toArray(new Requirement[0]);
        //reqCahceMap.put(locQId, result);
        Arrays.sort(res, reqcomparator);
        return res;
    }
    private static Comparator<Requirement> reqcomparator = new Comparator<Requirement>(){
		@Override
		public int compare(Requirement arg0, Requirement arg1) {
			return arg0.getUserFriendlyName().compareTo(arg1.getUserFriendlyName());
		}
    };
    /**
     * Returns document location is related to. If there is no such document - returns null. If there are two or 
     * more documents - returns first available and prints warning to console 
     * @return document related to location or null
     */
    public Document getDocument() {
    	return (Document)getParent();
    }
    
    public Document[] getDocuments() {
        UUID locQId = getUUId();
        return getDocumentsFor(getTreeDB(), locQId);
    }

    public static Document[] getDocumentsFor(TreeDB db, UUID locQId) {
        ArrayList<Document> result = new ArrayList<Document>();
        for (TreeNode node : new DeepFirstTreeWalker(db.getNode(Document.getTypeRootQId()))) {
            if (!node.getType().equals(Document.TYPE_NAME)) {
                continue;
            }
            Document doc = (Document) node;
            if (doc.getLocations().contains(locQId)) {
                result.add(doc);
            }
        }
        return result.toArray(new Document[0]);
    }

    private static String toString(TreeDB db, String locQId) {
        Requirement reqs[] = getRequirementsFor(db,locQId);
        if (reqs.length == 0) {
            return "orphan location(" + locQId + ")";
        }
        String[] names = new String[reqs.length];

        int i = 0;

        for (Requirement curReq : reqs) {
            String reqName = curReq.getUserFriendlyName(true);
            List<String> locQIds = curReq.getLocationQIds();
            if (locQIds.size() > 1) {
                names[i] = reqName + "#" + (locQIds.indexOf(locQId) + 1);
            } else {
                names[i] = reqName;
            }
            i++;
        }

        StringBuilder sb = new StringBuilder();
        sb.append(names[0]);
        for (i = 1; i < names.length; i++) {
            sb.append(", ");
            sb.append(names[i]);
        }
        return sb.toString();
    }

    @Override
    public String toString() {
        if(label==null) {
            label = Location.toString(getTreeDB(), getQualifiedId());
        }
        return label;
    }


    protected static final String ATTR_HIDDEN = "_hidden";

    public boolean isHidden() {
        return getBooleanAttributeValue( ATTR_HIDDEN );
    }

    public void setHidden(boolean hidden) {
        putAttribute(new Attribute(getUUId(), AttributeType.BOOL, ATTR_HIDDEN,hidden));
    }

	public void resetLabel() {
		if(!reqCahceMap.containsKey(getTreeDB()))
			reqCahceMap.put(getTreeDB(), new HashMap<String, Set<Requirement>>());
		String qid = getQualifiedId();
		if(reqCahceMap.get(getTreeDB()).containsKey(qid)){
			for(Requirement r:new HashSet<Requirement>(reqCahceMap.get(getTreeDB()).get(qid)))
				if(r.isDisposed())
					reqCahceMap.get(getTreeDB()).get(qid).remove(r);
			//if(reqCahceMap.get(qid).size()==0)
			//TODO look into situation
				reqCahceMap.get(getTreeDB()).remove(qid);
		}
		label = null;
		getTreeDB().fireNodeChange(this, true);
		getTreeDB().fireNodeGUIChange(this, true);
	}
	
	public int getIndexInDocument(){
		Document d = getDocument();
		if(d!=null)
			return d.getLocationIndex(getUUId());
		return -1;
	}
	
	public int compareIndexTo(Location l2) {
		Document doc1=getDocument(),
				 doc2=l2.getDocument();
		if(doc1==null || doc2==null)
			return 0;
		if(!doc1.getUUId().equals(doc2.getUUId()))
			return doc1.compareTo(doc2);
		return ((Integer)getIndexInDocument()).compareTo(l2.getIndexInDocument());
	}
	public static void clearCache(TreeDB db) {
		reqCahceMap.remove(db);
	}
	
//	public void setTransferredLocation() {
//		isTransferred = true;
//	}
//	
//	public boolean isTransferred() {
//		return isTransferred;
//		
//	}
}
