/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.handlers.location.CommonRemoveLocation;
import com.unitesk.requality.tools.Index;

public class Requirement extends TreeNode implements Comparable<TreeNode> {
    public static final String TYPE_NAME = "Requirement";
    public static final String ATTR_LOCATIONS = "_locations";
    private static final String TYPE_ROOT_QID = "/Requirements";
    public static final String ATTR_STATUS = "coverageStatus";
    public static final String ATTR_TRANSF_STATUS = "_transferStatus";
    public static final String ATTR_ORDER = "_order";
    
    public enum TransferStatus {
    	FAILED,
    	INCOMPLETE,
    	COMPLETE;
    	
    	public boolean isBetterThan(TransferStatus other) {
    		return this.ordinal() > other.ordinal();
    	}
    }
    
    public static String getTypeRootQId() {
        return TYPE_ROOT_QID;
    }

    
    
    // Constructor for eclipse instancer
    public Requirement() {
        this(null, null);
    }

    public Requirement(TreeDB db, UUID uuid) {
        super(db, uuid);
        setType(TYPE_NAME);
    }

    @SuppressWarnings("unchecked")
    public List<String> getLocationQIds() {
        Object val = getListAttributeValue(ATTR_LOCATIONS, AttributeType.STRING);
        if (val == null) {
            return Collections.emptyList();
        }
        if (val instanceof List) {
            return Collections.unmodifiableList((List<String>) val);
        }
        throw new IllegalStateException(ATTR_LOCATIONS
                + " value must be string list");
    }

	public List<Location> getLocations() {
        List<Location> result = new ArrayList<Location>();
        for (String locQId : getLocationQIds()) {
            TreeNode locNode = getTreeDB().getNode(locQId);
            if (locNode != null && locNode instanceof Location) {
                result.add((Location) locNode);
            }
        }
        return Collections.unmodifiableList(result);
    }

    public Location getLocation(String QId) {
        if (getLocationQIds().contains(QId)) {
            return (Location) getTreeDB().getNode(QId);
        }
        return null;
    }
    
    

    @SuppressWarnings("unchecked")
    public void addLocation(String nodeQId) {
        List<String> locations = (List<String>) getAttributeValue(ATTR_LOCATIONS);
        if (locations == null) {
            locations = new ArrayList<String>();
        }
        locations.add(nodeQId);
        putAttribute(new Attribute(getUUId(), AttributeType.LIST, ATTR_LOCATIONS,
                locations));
        // destroy locations label cache
        for (String l : locations) {
            Location loc = (Location) getTreeDB().getNode(l);
            if (loc != null) { // Can be if node not created still
                loc.resetLabel();
            }
        }
    }

    @SuppressWarnings("unchecked")
    public boolean removeLocation(String nodeQId) {
        Location loc = (Location) getTreeDB().getNode(nodeQId);
        if (loc != null) { // Can be if node not created still
            loc.resetLabel();
        }
        List<String> locations = (List<String>) getAttributeValue(ATTR_LOCATIONS);
        if (locations == null || !locations.remove(nodeQId)) {
            return false;
        }
        if (locations.size() > 0) {
            putAttribute(new Attribute(this, AttributeType.LIST, ATTR_LOCATIONS,
                    locations));
        } else {
            removeAttribute(ATTR_LOCATIONS);
        }
        // destroy locations label cache
        updateLocations(this);
        return true;
    }


    @Override
    public void setId(String id) {
        if (getQualifiedId().equals(Requirement.getTypeRootQId())) {
            throw new UnsupportedOperationException(
                    "Can't rename root requirement");
        }
        super.setId(id);
        //fire id update actions
    }
    @Override
    public void updateTreeNodes(TreeNode req, String what) {
        if(req.getType().equals(TYPE_NAME))
            updateLocations((Requirement)req);
        super.updateTreeNodes(req, what);
    }

    public void updateLocations(Requirement req) {
        for (String l : req.getLocationQIds()) {
            Location loc = (Location) getTreeDB().getNode(l);
            if (loc != null) { // Can be if node not created still
                loc.resetLabel();
            }
        }
    }

    public boolean isNameDefined() {
        return getAttributeValue(ATTR_NAME) != null;
    }

    @Override
    public int compareTo(TreeNode o) {
        if (o instanceof Requirement) {
            Requirement req = (Requirement) o;
            return getNameOrId().compareTo(req.getNameOrId());
        } else if (o instanceof TreeNode) {
            TreeNode req = (TreeNode) o;
            return getNameOrId().compareTo(req.getId());
        }
        throw new IllegalArgumentException();
    }

    public static String getFullName(TreeNode node, TreeDB db) {
        if (node.getParent() == null) {
            return "/";
        }
        String parentFullName = Requirement.getFullName(node.getParent(), db);
        String name;
        if(node instanceof Requirement) {
            name = ((Requirement) node).getNameOrId();
        } else {
            name = node.getId();
        }
        if (!parentFullName.equals("/")) {
            parentFullName = parentFullName + "/";
        }
        return parentFullName + name;
    }

    @Override
    public String getUserFriendlyName(boolean inner) {
//        if (getId().equals("Requirements"))
  //          return "/"+inner?getId():"";
        if (getName() != null)
            return getName();
        return super.getUserFriendlyName(inner);
    }

	public int compareLocationIndexes(Requirement r){
		int min1=0,min2=0;
		Location[]loc1=getLocations().toArray(new Location[0]),
				  loc2=r.getLocations().toArray(new Location[0]);
		for(int i = 1;i<loc1.length;i++)
			if(loc1[i].compareIndexTo(loc1[min1])<0)
				min1=i;
		for(int i = 1;i<loc2.length;i++)
			if(loc2[i].compareIndexTo(loc2[min2])<0)
				min2=i;
		if(loc1.length == 0 || loc2.length == 0)
			return ((Integer)loc1.length).compareTo(loc2.length);
		return loc1[min1].compareIndexTo(loc2[min2]);
	}
	
	public void setTransferProblem(TransferStatus status) {
		putAttribute(new Attribute(this, AttributeType.STRING, ATTR_TRANSF_STATUS, status.toString()));
		saveAttributes();
	}
}
